import matplotlib

matplotlib.use("Agg")


from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RattleGenerator
import shephex

@shephex.chain()
def raffle_agox(train=True, n_dist=4, index=42):

    # Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
    # from argparse import ArgumentParser
    # parser = ArgumentParser()
    # parser.add_argument('-i', '--run_idx', type=int, default=0)
    # args = parser.parse_args()

    # seed = args.run_idx
    # database_index = args.run_idx

    ##############################################################################
    # System & general settings:
    ##############################################################################

    import numpy as np
    from ase import Atoms
    from ase.io import write
    from ase.calculators.singlepoint import SinglePointCalculator
    from agox.generators.raffle import RaffleGenerator, bounds_to_confinement
    from mace.calculators import mace_mp
    from agox.utils.replica_exchange.priors import get_prior


    # Manually set seed and database-index
    seed = index
    database_index = index
    np.random.seed(seed)

    chg = get_prior("chg")
    mace = get_prior("mace")

    from ase.build import bulk
    from ase.io import write, read

    from ase.build import bulk
    from ase.io import write
    cell = np.eye(3) * 10
    template = bulk('C', crystalstructure="diamond", orthorhombic=True)
    template=template.repeat((2,2,4))
    template.calc = mace
    template_copy = template.copy()

    C_reference_energy = template.get_potential_energy() / len(template)

    indices = 0

    for i in range(len(template))[::-1]:
        if template[i].position[2] < template.cell[2][2]/2:
            del template[i]
            indices += 1

    #for i in range(len(template))[::-1]:
    #    if i > 4:
    #        del template[i]
    #        indices += 1
    #write("template.traj", template)
    print(indices)

    confinement_corner, confinement_cell = bounds_to_confinement([[0, 0, 0], [1, 1, 1]],template)

    element_energies = {'C': C_reference_energy}
    species_to_place={"C" : indices}

    symbols = ""
    for key in species_to_place.keys():
        symbols += key
        symbols += str(species_to_place[key])

    environment = Environment(
        template=template,
        symbols=symbols,
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
    )
    print("Environment: Setup")
    # Database
    db_path = f"../db{database_index}_dist_{n_dist}.db"  # From input argument!
    database = Database(filename=db_path, order=5)
    database.restore_to_memory()


    generator = RaffleGenerator(**environment.get_confinement(),
        database=database,
        environment=environment,
        element_energies=element_energies,
        transfer_data=[template_copy],
        order=1,
        n_dist=n_dist,
        n_structures=1,
        sampler=None,
        train=train,
        seed=seed,
        method_ratio = {"void": 0.0, "rand": 0.1, "walk": 0.25, "grow": 0.25, "min": 1.0})

    ##############################################################################
    # Search Settings:
    ##############################################################################

    # Number of steps is very low - should be set higher for a real search!
    #relaxer = ParallelRelaxPostprocess(mace,
    #                                   optimizer=BFGS,
    #                                   order=2,
    #                                   optimizer_run_kwargs={"steps": 100, "fmax": 0.05},
    #                                   fix_template=True)


    # With the model pre-relax we dont want to take many steps in the real potential!
    # As we are training the model all the data is saved with the store_trajectory argument.
    evaluator = LocalOptimizationEvaluator(
        mace,
        gets={"get_key": "candidates"},
        store_trajectory=False,
        optimizer_run_kwargs={"fmax": 0.05, "steps": 100},
        order=3,
        number_to_evaluate=1,
        constraints=environment.get_constraints(),
        fix_template = True,
    )
    from agox.postprocessors.minimum_dist import MinimumDistPostProcess

    c1 = 0.7
    c2 = 3
    minimum_dist = MinimumDistPostProcess(
        c1=c1,
        c2=c2,
        order=1.5,
    )
    minimum_dist2 = MinimumDistPostProcess(
        c1=0.9,
        gets = {"get_key" : "evaluated_candidates"},
        sets = {"set_key" : "evaluated_candidates"},
        c2=c2,
        order=3.5,
    )

    ##############################################################################
    # Let get the show running!
    ##############################################################################

    agox = AGOX(generator, minimum_dist, minimum_dist2, database, evaluator, seed=seed)

    agox.run(N_iterations=1000)

if __name__ == "__main__":

    walkers = 10

    experiments = (raffle_agox(directory='n-body/').
                   zip(train=[True]).
                   permute(index=range(50)).
                   permute(n_dist=[2,3,4]))

    executor = shephex.executor.SlurmExecutor(ntasks=walkers, nodes=1,
            time='1-00:00:00',
            mem_per_cpu='10G',
            partition='q48,q40,q36',
            job_name="AGOX-RAFFLE",
            directory="submission",
            safety_check = False,
            scratch=False)

    executor.execute(experiments)



from chgnet.model import CHGNetCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build, Atoms
from ase.optimize import FIRE
from ase.io import write
import numpy as np
import os
from copy import deepcopy
from joblib import Parallel, delayed

import logging
logging.basicConfig(level=logging.DEBUG)

# Function to relax a structure
def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration):
    # Check if the structure has already been processed
    if i < num_structures_old:
        return
    
    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None
    
    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None
    
    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None
    
    return atoms, energy_unrlxd, energy_rlxd


if __name__ == "__main__":

    # set up the calculator
    calc_params = {}
    calc = CHGNetCalculator()

    # set up the hosts
    crystal_structures = [
        'orthorhombic', 'hcp',
    ]
    hosts = []
    for crystal_structure in crystal_structures:
        print(f'Crystal structure: {crystal_structure}')
        for a in np.linspace(3.1, 5.4, num=6):
            b = a
            c = a
            atom = build.bulk(
                    name = 'Al',
                    crystalstructure = crystal_structure,
                    a = a,
                    b = b,
                    c = c,
            )
            hosts.append(Atoms('Al', positions=[(0, 0, 0)], cell=atom.get_cell(), pbc=True, calculator=calc))
    print("number of hosts: ", len(hosts))

    # set the parameters for the generator
    optimise_structure = True
    mass = 26.9815385
    density = 1.61 # u/A^3

    # loop over random seeds
    for seed in range(1):
        print(f"Seed: {seed}")
        energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
        energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
        generator = raffle_generator()
        generator.distributions.set_element_energies(
            {
                'Al': 0.0
            }
        )
        # set energy scale
        generator.distributions.set_kBT(0.4)
        # set the distribution function widths (2-body, 3-body, 4-body)
        generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

        # set the initial database
        initial_database = [Atoms('Al', positions=[(0, 0, 0)], cell=[8, 8, 8], pbc=True)]
        initial_database[0].calc = calc
        generator.distributions.create(initial_database)

        # check if the energies file exists, if not create it
        if os.path.exists(energies_rlxd_filename):
            with open(energies_rlxd_filename, "w") as energy_file:
                pass
        else:
            open(energies_rlxd_filename, "w").close()

        if os.path.exists(energies_unrlxd_filename):
            with open(energies_unrlxd_filename, "w") as energy_file:
                pass
        else:
            open(energies_unrlxd_filename, "w").close()

        # initialise the number of structures generated
        num_structures_old = 0
        unrlxd_structures = []
        rlxd_structures = []
        iter2 = 0
        generator.init_seed(seed)
        # start the iterations, loop over the hosts, then repeat the process X times
        for iter in range(10):
            for host in hosts:
                generator.set_host(host)
                volume = host.get_volume()

                # calculate the number of atoms in the host to achieve the desired density
                num_atoms = round(density * volume / mass) - 1
                if(num_atoms < 1):
                    continue
                iter2 += 1
                print(f"Volume: {volume}")
                print(f"Number of atoms: {num_atoms}")
    
                # generate the structures
                generator.generate(
                    num_structures = 5,
                    stoichiometry = { 'Al': num_atoms },
                    method_ratio = {"void": 0.1, "rand": 0.01, "walk": 0.25, "grow": 0.25, "min": 1.0},
                    verbose = 0,
                )

                # print the number of structures generated
                print("Total number of structures generated: ", generator.num_structures)
                generated_structures = generator.get_structures(calc)
                num_structures_new = len(generated_structures)

                # check if directory iteration[iter] exists, if not create it
                iterdir = f"iteration{iter2}/"
                if not os.path.exists(iterdir):
                    os.makedirs(iterdir)
                generator.print_settings(iterdir+"generator_settings.txt")

                # set up list of energies
                energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
                energy_rlxd = np.zeros(num_structures_new - num_structures_old)
                for i in range(num_structures_new - num_structures_old):
                    write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
                    print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
                    unrlxd_structures.append(generated_structures[num_structures_old + i].copy())
                    unrlxd_structures[-1].calc = SinglePointCalculator(
                        generated_structures[num_structures_old + i],
                        energy=generated_structures[num_structures_old + i].get_potential_energy(),
                        forces=generated_structures[num_structures_old + i].get_forces()
                    )
                
                # Start parallel execution
                print("Starting parallel execution")
                results = Parallel(n_jobs=5)(
                    delayed(process_structure)(i, generated_structures[i].copy(), num_structures_old, calc_params, optimise_structure, iteration=seed)
                    for i in range(num_structures_old, num_structures_new)
                )

                # Wait for all futures to complete
                for j, result in enumerate(results):
                    generated_structures[num_structures_old + j], energy_unrlxd[j], energy_rlxd[j] = result
                    if generated_structures[num_structures_old + j] is None:
                        print("Structure failed the checks")
                        continue
                    rlxd_structures.append(generated_structures[num_structures_old + j].copy())
                    rlxd_structures[-1].calc = SinglePointCalculator(
                        generated_structures[j+num_structures_old],
                        energy=generated_structures[num_structures_old + j].get_potential_energy(),
                        forces=generated_structures[num_structures_old + j].get_forces()
                    )
                print("All futures completed")

                # Remove structures that failed the checks
                for j, atoms in reversed(list(enumerate(generated_structures))):
                    if atoms is None:
                        energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                        energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                        del generated_structures[j]
                        del unrlxd_structures[j]
                        del rlxd_structures[j]
                        generator.remove_structure(j)
                num_structures_new = len(generated_structures) 

                # write the structures to files
                for i in range(num_structures_new - num_structures_old):
                    write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
                    print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
                    # append energy per atom to the 'energies_unrlxd_filename' file
                    with open(energies_unrlxd_filename, "a") as energy_file:
                        energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
                    # append energy per atom to the 'energies_rlxd_filename' file
                    with open(energies_rlxd_filename, "a") as energy_file:
                        energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

                # update the distribution functions
                print("Updating distributions")
                generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)

                # print the new distribution functions to a file
                print("Printing distributions")
                generator.distributions.write_dfs(iterdir+"distributions.txt")
                generator.distributions.write_2body(iterdir+"df2.txt")
                generator.distributions.write_3body(iterdir+"df3.txt")
                generator.distributions.write_4body(iterdir+"df4.txt")
                generator.distributions.deallocate_systems()

                # update the number of structures generated
                num_structures_old = num_structures_new

        # Write the final distribution functions to a file
        generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

        # Read energies from the file
        with open(energies_rlxd_filename, "r") as energy_file:
            energies = energy_file.readlines()

        # Parse and sort the energies
        energies = [line.strip().split() for line in energies]
        energies = sorted(energies, key=lambda x: float(x[1]))

        # Write the sorted energies back to the file
        with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
            for entry in energies:
                energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

        # Write the structures to files
        write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
        write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
        print("All generated and relaxed structures written")

    print("Learning complete")


import os

def pytest_addoption(parser):
    parser.addoption(
        "--fortran-compiler", 
        action="store", 
        default="gfortran",  # Default compiler
        help="Specify the Fortran compiler to use"
    )

def pytest_configure(config):
    # Make the Fortran compiler available globally during tests
    compiler = config.getoption("--fortran-compiler")
    os.environ["FORTRAN_COMPILER"] = compiler


import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms
from ase.optimize import BFGS

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RattleGenerator
from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.postprocessors.ray_relax import ParallelRelaxPostprocess
from agox.samplers import MetropolisSampler
from ase import build
from pathlib import Path

script_dir = Path(__file__).resolve().parent

def rss_agox(index=42):
    # Manually set seed and database-index
    seed = index
    database_index = index
    np.random.seed(seed)

    ##############################################################################
    # System & general settings:
    ##############################################################################

    from ase import Atoms
    from ase.io import write
    from ase.calculators.singlepoint import SinglePointCalculator
    from agox.generators import RandomGenerator
    from agox.generators.raffle import bounds_to_confinement
    from mace.calculators import mace_mp
    import os


    # check if mace file exists
    if not os.path.exists(script_dir  / ".." / ".." / "mace-mpa-0-medium.model"):
        print("MACE-MPA-0 model file not found. Please download the model from the MACE website.")
        print("https://github.com/ACEsuit/mace-foundations/releases/tag/mace_mpa_0")
        exit(1)

    # set up the calculator
    # calc_params = {}
    # calc = CHGNetCalculator()
    calc_params = { 'model':  script_dir/ ".." / ".." / "mace-mpa-0-medium.model" }
    mace = mace_mp(**calc_params)

    from ase.build import bulk
    from ase.io import write, read

    template = read(script_dir / ".." / "POSCAR_host_gb")

    bounds = [[0.20, 0, 0.18], [0.30, 1, 0.33]]
    confinement_corner, confinement_cell  = bounds_to_confinement(bounds, template)

    environment = Environment(
        template=template,
        symbols="C15",
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
    )
    print("Environment: Setup")
    # Database
    db_path = "../db{}.db".format(database_index)  # From input argument!
    database = Database(filename=db_path, order=5)
    sampler = MetropolisSampler(temperature=0.25, order=4)

    generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

    print("Generator: Setup")
    ##############################################################################
    # Search Settings:
    ##############################################################################

    # With the model pre-relax we dont want to take many steps in the real potential!
    # As we are training the model all the data is saved with the store_trajectory argument.
    evaluator = LocalOptimizationEvaluator(
        mace,
        gets={"get_key": "candidates"},
        store_trajectory=False,
        optimizer_run_kwargs={"fmax": 0.05, "steps": 200},
        order=3,
        constraints=environment.get_constraints(),
    )
    from agox.postprocessors.minimum_dist import MinimumDistPostProcess

    c1 = 0.8
    c2 = 3  
    minimum_dist = MinimumDistPostProcess(
        c1=c1,
        c2=c2,
        order=2.5,
    )
    ##############################################################################
    # Let get the show running!
    ##############################################################################

    agox = AGOX(generator, minimum_dist, minimum_dist, database, evaluator, seed=seed)

    agox.run(N_iterations=1000)

    structure_data = database.get_all_structures_data()
    # check iteration number has changed
    iteration_check = -1
    unrlxd_structures = []
    rlxd_structures = []
    for idx, structure in enumerate(structure_data):
        if structure["iteration"] != iteration_check:
            iteration_check = structure["iteration"]
            unrlxd_structures.append(database.db_to_atoms(structure))
            if idx != 0:
                rlxd_structures.append(database.db_to_atoms(structure_data[idx-1]))
    if len(rlxd_structures) != len(unrlxd_structures):
        rlxd_structures.append(database.db_to_atoms(structure_data[-1]))

    print("Unrelaxed structures", len(unrlxd_structures))
    print("Relaxed structures", len(rlxd_structures))
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)


if __name__ == "__main__": 
    
    rss_agox()


# This script tidies up the n-body database files by organizing them into subdirectories based on the number of bodies.
# By running this script, then the following agox line in the command line:
#    agox analysis method_ratios/*
# you will get analysis of the success split up by the method ratios used in the generation of the structures.

import os
import re
import shutil
from math import gcd
from functools import reduce

src_dir = "method_ratio"
dst_base = "method_ratios"

# Regex to match the filename pattern
pattern = re.compile(
    r"^db(?P<seed>\d+)_dist_grow(?P<grow>[\d\.]+)_min(?P<min>[\d\.]+)_rand(?P<rand>[\d\.]+)_void(?P<void>[\d\.]+)_walk(?P<walk>[\d\.]+)\.db$"
)

def float_str_to_int_ratios(values):
    # Convert all to floats
    floats = [float(v) for v in values]
    scale = 1000  # To avoid float precision issues
    ints = [round(f * scale) for f in floats]
    divisor = reduce(gcd, ints)
    return [i // divisor for i in ints]

os.makedirs(dst_base, exist_ok=True)

for fname in os.listdir(src_dir):
    match = pattern.match(fname)
    if match:
        parts = match.groupdict()
        ratios = float_str_to_int_ratios([
            parts["void"], parts["rand"], parts["walk"], parts["grow"], parts["min"]
        ])
        # Assign to the appropriate method
        v, r, w, g, m = ratios
        # exit()
        subdir_name = f"v{v}_r{r}_w{w}_g{g}_m{m}"
        fname_new = f"db{parts['seed']}.db" #_{subdir_name}.db"
        subdir_path = os.path.join(dst_base, subdir_name)
        os.makedirs(subdir_path, exist_ok=True)

        src_path = os.path.join(src_dir, fname)
        dst_path = os.path.join(subdir_path, fname_new)
        shutil.copy2(src_path, dst_path)



from chgnet.model import CHGNetCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build, Atoms
from ase.optimize import FIRE
from ase.io import write
import numpy as np
import os
from joblib import Parallel, delayed

import logging
logging.basicConfig(level=logging.DEBUG)

# Function to relax a structure
def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration, calc):
    # Check if the structure has already been processed
    if i < num_structures_old:
        return None, None, None

    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None

    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None

    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None

    return atoms, energy_unrlxd, energy_rlxd


if __name__ == "__main__":

    # set up the calculator
    calc_params = {}
    calc = CHGNetCalculator()


    cell = [
        [1.22799768, -2.12695438, 0.00000000],
        [1.22799768, 2.12695438, 0.00000000],
        [0.00000000, 0.00000000, 7.31932291]
    ]
    host = Atoms('C', positions=[(0, 0, 0)], cell=cell, pbc=True, calculator=calc)


    # set the parameters for the generator
    optimise_structure = True
    mass = 12.011
    num_atoms = 7


    seed = 0
    energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
    energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
    generator = raffle_generator()
    generator.distributions.set_history_len(10)
    generator.distributions.set_element_energies(
        {
            'C': 0.0
        }
    )

    # set energy scale
    generator.distributions.set_kBT(0.2)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

    # set the initial database
    initial_database = [Atoms('C', positions=[(0, 0, 0)], cell=[8, 8, 8], pbc=True)]
    initial_database[0].calc = calc
    generator.distributions.create(initial_database)

    # set the host
    generator.set_host(host)

    # check if the energies file exists, if not create it
    if os.path.exists(energies_rlxd_filename):
        with open(energies_rlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_rlxd_filename, "w").close()

    if os.path.exists(energies_unrlxd_filename):
        with open(energies_unrlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_unrlxd_filename, "w").close()

    # initialise the number of structures generated
    num_structures_old = 0
    unrlxd_structures = []
    rlxd_structures = []
    # start the iterations
    generator.init_seed(put=seed)
    generator.set_method_ratio_default({"void": 0.1, "rand": 0.01, "walk": 0.25, "grow": 0.25, "min": 1.0})
    for iter in range(200):

        # generate the structures
        generator.generate(
            num_structures = 5,
            stoichiometry = { 'C': 3 },
            verbose = 1,
        )

        # print the number of structures generated
        print("Total number of structures generated: ", generator.num_structures)
        generated_structures = generator.get_structures(calc)
        num_structures_new = len(generated_structures)

        # check if directory iteration[iter] exists, if not create it
        iterdir = f"iteration{iter}/"
        if not os.path.exists(iterdir):
            os.makedirs(iterdir)
        generator.print_settings(iterdir+"generator_settings.txt")

        # set up list of energies
        energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
        energy_rlxd = np.zeros(num_structures_new - num_structures_old)
        for i in range(num_structures_new - num_structures_old):
            write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
            print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
            unrlxd_structures.append(generated_structures[num_structures_old + i].copy())
            unrlxd_structures[-1].calc = SinglePointCalculator(
                generated_structures[num_structures_old + i],
                energy=generated_structures[num_structures_old + i].get_potential_energy(),
                forces=generated_structures[num_structures_old + i].get_forces()
            )

        # Start parallel execution
        print("Starting parallel execution")
        results = Parallel(n_jobs=5)(
            delayed(process_structure)(i, generated_structures[i].copy(), num_structures_old, calc_params, optimise_structure, iteration=seed, calc=calc)
            for i in range(num_structures_old, num_structures_new)
        )

        # Wait for all futures to complete
        for j, result in enumerate(results):
            generated_structures[num_structures_old + j], energy_unrlxd[j], energy_rlxd[j] = result
            if generated_structures[num_structures_old + j] is None:
                print("Structure failed the checks")
                continue
            rlxd_structures.append(generated_structures[num_structures_old + j].copy())
            rlxd_structures[-1].calc = SinglePointCalculator(
                generated_structures[j+num_structures_old],
                energy=generated_structures[num_structures_old + j].get_potential_energy(),
                forces=generated_structures[num_structures_old + j].get_forces()
            )
        print("All futures completed")

        # Remove structures that failed the checks
        for j, atoms in reversed(list(enumerate(generated_structures))):
            if atoms is None:
                energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                del generated_structures[j]
                del unrlxd_structures[j]
                del rlxd_structures[j]
                generator.remove_structure(j)
        num_structures_new = len(generated_structures)

        # write the structures to files
        for i in range(num_structures_new - num_structures_old):
            write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
            print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
            # append energy per atom to the 'energies_unrlxd_filename' file
            with open(energies_unrlxd_filename, "a") as energy_file:
                energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
            # append energy per atom to the 'energies_rlxd_filename' file
            with open(energies_rlxd_filename, "a") as energy_file:
                energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

        # update the distribution functions
        print("Updating distributions")
        generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)
        print("deltas", generator.distributions.history_deltas)
        if generator.distributions.is_converged(threshold=1e-3):
            print("Convergence reached")

        # print the new distribution functions to a file
        print("Printing distributions")
        generator.distributions.write_dfs(iterdir+"distributions.txt")
        generator.distributions.write_2body(iterdir+"df2.txt")
        generator.distributions.write_3body(iterdir+"df3.txt")
        generator.distributions.write_4body(iterdir+"df4.txt")
        generator.distributions.deallocate_systems()

        # update the number of structures generated
        num_structures_old = num_structures_new

    # Write the final distribution functions to a file
    generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

    # Read energies from the file
    with open(energies_rlxd_filename, "r") as energy_file:
        energies = energy_file.readlines()

    # Parse and sort the energies
    energies = [line.strip().split() for line in energies]
    energies = sorted(energies, key=lambda x: float(x[1]))

    # Write the sorted energies back to the file
    with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
        for entry in energies:
            energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

    # Write the structures to files
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
    print("All generated and relaxed structures written")

    print("Learning complete")



# Configuration file for the Sphinx documentation builder.

# -- Project information
import datetime
import os
import sys

from unittest.mock import Mock

MOCK_MODULES = ["raffle._raffle"]  # List any other modules if needed
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

# sys.path.insert(0, os.path.abspath(os.path.join('..', '..', 'src', 'raffle')))  # Sets the base path to find your modules
sys.path.insert(0, os.path.abspath(os.path.join('..', '..', 'src')))  # Sets the base path to find your modules

project = 'RAFFLE'
copyright = f'{datetime.date.today().year}, RAFFLE-developers'
# release = '1.0'
# version = '1.0.0'

# -- General configuration
master_doc = 'index'

# Identify the branch of the documentation
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    git_branch = os.environ.get("READTHEDOCS_GIT_IDENTIFIER", "main")
else:
    git_branch = "main"  # or get from git directly with subprocess

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinxcontrib.bibtex',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_rtd_theme',
    'sphinx.ext.extlinks',
]

extlinks = {
    'git': ('https://github.com/ExeQuantCode/RAFFLE/blob/' + git_branch + '/%s', 'git: %s')
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

exclude_patterns = ['_build', '.DS_Store', 'build']


# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'

html_logo = "RAFFLE_logo_no_background.png"
# html_favicon = 'favicon.ico'
html_theme_options = {
    'logo_only': False,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    # 'style_nav_header_background': 'white',
    'flyout_display': 'hidden',
    'version_selector': True,
    'language_selector': True,
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False,
}


html_context = {
    "display_github": True,
    "github_repo": "RAFFLE",
    "github_user": "ExeQuantCode",
    "github_version": "main",
    "conf_py_path": "/docs/source/",
}

autoclass_content="both"

bibtex_bibfiles = ['references.bib']



import os
nthreads = 1
os.environ["OMP_NUM_THREADS"] = str(nthreads)
os.environ["OPENBLAS_NUM_THREADS"] = str(nthreads)
os.environ["MKL_NUM_THREADS"] = str(nthreads)

import pytest
import time
import numpy as np
from ase import Atoms
from ase.io import read
from chgnet.model import CHGNetCalculator
from raffle.generator import raffle_generator


@pytest.mark.parametrize("kBT", [0.4])
@pytest.mark.parametrize("width", [
    [0.025, np.pi/200.0, np.pi/200.0]
])
@pytest.mark.parametrize("grid_spacing", [0.1])
@pytest.mark.parametrize("method_ratio", [
  {'void': 0.0, 'rand': 0.0, 'walk': 0.0, 'grow': 0.0, 'min': 1.0}
])
@pytest.mark.parametrize("cutoff_max", [4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0])
@pytest.mark.benchmark(
    group="cutoff_in_evaluator",
    min_rounds=5,
    timer=time.time,
    disable_gc=True,
    warmup=False
)
def test_cutoff_in_evaluator(benchmark, grid_spacing, kBT, width, method_ratio, cutoff_max):
    
    calc = CHGNetCalculator()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(current_dir, '..', '..', '..', 'example', 'data')
    filename = os.path.join(data_dir, 'diamond.xyz')
    atoms = read(filename, index=":")


    # Initialise generator
    generator = raffle_generator()

    # set energy scale
    generator.distributions.set_kBT(kBT)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width(width)

    # get list of elements in list of atoms
    elements = []
    for atom in atoms:
        elements += atom.get_chemical_symbols()
    elements = list(set(elements))
    energies = {element: 0.0 for element in elements}
    generator.distributions.set_element_energies(energies)

    # set the cutoff_max for the distributions
    generator.distributions.set_cutoff_max([cutoff_max, np.pi, np.pi])

    generator.set_grid(grid_spacing=grid_spacing)

    generator.distributions.create(atoms)
    
    benchmark.extra_info = {
        "kBT": kBT,
        "width": width,
        "grid_spacing": grid_spacing,
        "method_ratio": method_ratio,
        "cutoff_max": cutoff_max
    }

    # Measure time for distributions.evaluate()
    result = benchmark(generator.evaluate, atoms[0])

    return result


@pytest.mark.parametrize("kBT", [0.4])
@pytest.mark.parametrize("width", [
    [0.025, np.pi/200.0, np.pi/200.0]
])
@pytest.mark.parametrize("grid_spacing", [0.1])
@pytest.mark.parametrize("method_ratio", [
  {'void': 0.0, 'rand': 0.0, 'walk': 0.0, 'grow': 0.0, 'min': 1.0}
])
@pytest.mark.parametrize("cutoff_max", [4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0])
@pytest.mark.benchmark(
    group="cutoff_in_database",
    min_rounds=5,
    timer=time.time,
    disable_gc=True,
    warmup=False
)
def test_cutoff_in_database(benchmark, grid_spacing, kBT, width, method_ratio, cutoff_max):
    
    calc = CHGNetCalculator()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(current_dir, '..', '..', '..', 'example', 'data')
    filename = os.path.join(data_dir, 'diamond.xyz')
    atoms = read(filename, index=":")


    # Initialise generator
    generator = raffle_generator()

    # set energy scale
    generator.distributions.set_kBT(kBT)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width(width)

    # get list of elements in list of atoms
    elements = []
    for atom in atoms:
        elements += atom.get_chemical_symbols()
    elements = list(set(elements))
    energies = {element: 0.0 for element in elements}
    generator.distributions.set_element_energies(energies)

    # set the cutoff_max for the distributions
    generator.distributions.set_cutoff_max([cutoff_max, np.pi, np.pi])

    generator.set_grid(grid_spacing=grid_spacing)

    benchmark.extra_info = {
        "kBT": kBT,
        "width": width,
        "grid_spacing": grid_spacing,
        "method_ratio": method_ratio,
        "cutoff_max": cutoff_max
    }

    # Measure time for distributions.create()
    result = benchmark(generator.distributions.create, atoms)

    return result

if __name__ == '__main__':
    cutoff_max = [6.0]
    results = []

    for cutoff in cutoff_max:
        for _ in range(5):  # Run the benchmark 5 times for each cutoff
            result = test_cutoff_in_database(lambda x: x, 0.1, 0.4, [0.025, np.pi/200.0, np.pi/200.0], {'void': 0.0, 'rand': 0.0, 'walk': 0.0, 'grow': 0.0, 'min': 1.0}, cutoff)
    
    for cutoff in cutoff_max:
        for _ in range(5):  # Run the benchmark 5 times for each cutoff
            result = test_cutoff_in_evaluator(lambda x: x, 0.1, 0.4, [0.025, np.pi/200.0, np.pi/200.0], {'void': 0.0, 'rand': 0.0, 'walk': 0.0, 'grow': 0.0, 'min': 1.0}, cutoff)

    # Print results
    print("\nBenchmark Results:")
    print("-" * 40)
    print(f"{'Grid Size':<10} {'Time (s)':<10}")
    print("-" * 40)
    for result in results:
        print(f"{result['grid_spacing']:<10} {result['time']:<10.4f}")



from chgnet.model import CHGNetCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build, Atoms
from ase.optimize import FIRE
from ase.io import write, read
import numpy as np
import os
from joblib import Parallel, delayed
from pathlib import Path

script_dir = Path(__file__).resolve().parent


# Function to relax a structure
def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration):
    # Check if the structure has already been processed
    if i < num_structures_old:
        return None, None, None
    
    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None
    
    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None
    
    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None
    
    return atoms, energy_unrlxd, energy_rlxd


if __name__ == "__main__":

    # set up the calculator
    calc_params = {}
    calc = CHGNetCalculator()

    hosts = []
    # get all hosts in ../../data/C-MgO_hosts/ directory
    for filename in os.listdir(script_dir / ".." / ".." / ".." / "data" / "C-MgO_hosts"):
        if filename.startswith("POSCAR"):
            atoms = read(script_dir / ".." / ".." / ".." / "data" / "C-MgO_hosts" / filename)
            if not all(element == 'C' for element in set(atoms.get_chemical_symbols())):
                print(f"Skipping {filename} as it has non-carbon atoms")
                continue
            elif len(atoms) > 20:
                print(f"Skipping {filename} as it has more than 20 atoms (want to do smaller cells for now)")
                continue
            hosts.append(read(script_dir / ".." / ".." / ".." / "data" / "C-MgO_hosts" / filename))
            hosts[-1].calc = calc

    # set up element reference energies
    C_reference = Atoms('C4', positions=[
        [0.000000,    0.000000,    1.950768],
        [0.000000,    0.000000,    5.852305],
        [0.000000,    1.424491,    1.950768],
        [1.233646,    0.712246,    5.852305],
        ], cell=[[1.2336, -2.1367, 0.0], [1.2336, 2.1367, 0.0], [0.0, 0.0, 7.8031]], pbc=True)
    C_reference.calc = calc
    C_reference_energy = C_reference.get_potential_energy() / len(C_reference)
    Mg_reference = Atoms("Mg2", positions=[
            [0.000000,    1.831369,    1.285301],
            [1.586012,    0.915684,    3.855904],
        ], cell=[
            [1.58601,   -2.7471,    0.0000],
            [1.58601,    2.7471,    0.0000],
            [0.00000,    0.0001,    5.1412],
        ], pbc=True)
    Mg_reference.calc = calc
    Mg_reference_energy = Mg_reference.get_potential_energy() / len(Mg_reference)
    O_reference = Atoms("O2", positions=[
        [0.0, 0.0, 0.0],
        [1.23, 0.0, 0.0]], cell=[10, 10, 10], pbc=False)
    O_reference.calc = calc
    O_reference_energy = O_reference.get_potential_energy() / len(O_reference)

    generator = raffle_generator()
    generator.distributions.set_element_energies(
        {
            'C': C_reference_energy,
            'Mg': Mg_reference_energy,
            'O': O_reference_energy
        }
    )

    # set the covalent radii
    generator . distributions . set_bond_radii (
        {
            ('C', 'Mg'): 1.4,
            ('Mg', 'Mg'): 1.4,
            ('C', 'O'): 1.4,
            # ('C', 'Mg'): 1.15, # 2.3 = what Joe used https://github.com/ExeQuantCode/RAFFLE/blob/1cfdf2d220dce41f20ace4cce5a4135a37d1f1b1/chem.in
            # ('C', 'O'): 0.587,
            # ('Mg', 'Mg'): 1.592,
            # INITIALISE O2 BOND TO LARGER VALUE
        }
    )

    # set energy scale
    generator.distributions.set_kBT(0.4)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

    # generate MgO bulk
    MgO_atom = Atoms("Mg4O4", positions=[
            [0.0, 0.0, 0.0],
            [0.0, 2.097, 2.097],
            [2.097, 0.0, 2.097],
            [2.097, 2.097, 0.0],
            [0.0, 0.0, 2.097],
            [0.0, 2.097, 0.0],
            [2.097, 0.0, 0.0],
            [2.097, 2.097, 2.097],
        ], cell=[4.1940, 4.1940, 4.1940], pbc=True)
    MgO_atom.calc = calc

    # read C-Mg-O database, obtained from Materials Project
    print("Reading database")
    # database = read("../database.xyz", index=":")
    database = [ C_reference, Mg_reference, O_reference, MgO_atom ]
    generator.distributions.create(database, deallocate_systems=False)

    # # set up the initial database
    # initial_database = []
    # for i, atoms in enumerate(database):
    #     if atoms.calc is None:
    #         database.remove(atoms)
    #         continue
    #     # check if any other species in the system other than C, Mg, O
    #     if not all(element in ['C', 'Mg', 'O'] for element in set(atoms.get_chemical_symbols())):
    #         print(f"Skipping structure {i} as it has non-C-Mg-O atoms")
    #         continue
    #     atoms.calc = calc
    #     initial_database.append(atoms)
    # initial_database.append(C_reference)
    # initial_database.append(Mg_reference)
    # initial_database.append(O_reference)
    # initial_database.append(MgO_atom)
    # generator.distributions.create(initial_database, deallocate_systems=False)

    # set parameters
    seed = 0

    # check if the energies file exists, if not create it
    energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
    energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
    if os.path.exists(energies_rlxd_filename):
        with open(energies_rlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_rlxd_filename, "w").close()

    if os.path.exists(energies_unrlxd_filename):
        with open(energies_unrlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_unrlxd_filename, "w").close()
    
    # initialise the number of structures generated
    iter = -1
    num_structures_old = 0
    unrlxd_structures = []
    rlxd_structures = []
    optimise_structure = True
    mass_C = 12.011
    mass_Mg = 24.305
    mass_O = 15.999
    generator.init_seed(put=seed)
    # loop over all hosts
    for host in hosts:
        print("setting host")
        generator.set_host(host)
        volume = host.get_volume()
        num_atoms_C = len(host)
    
        host_positions = host.get_scaled_positions(wrap=True)
        host_height = np.linalg.norm(host.get_cell()[2, :])
        print(f"host_height: {host_height}")

        # get all unique z values
        z_values = np.unique(host_positions[:, 2])
        print(f"z_values: {z_values}")
        # return the two closes to 0.5 in the cell
        z_values = z_values[np.argsort(np.abs(z_values - 0.5))[:2]]
        print(f"z_values: {z_values}")
        # set the bounds for the generator
        z_min = min(z_values) + 1.0 / host_height
        z_max = max(z_values) - 1.0 / host_height
        generator.set_bounds([[0, 0, z_min], [1, 1, z_max]])

        # loop over range of MgO densities to intercalate
        for num_atoms in range(1, 50):

            num_atoms_Mg = num_atoms
            num_atoms_O = num_atoms
            density = ( mass_C * num_atoms_C + mass_Mg * num_atoms_Mg + mass_O * num_atoms_O ) / volume
            if density > 2.2:
                # print("Density too high:", density, "u/A^3")
                continue
            elif density < 1.0:
                # print("Density too low", density, "u/A^3")
                continue
            iter += 1

            # generate structures
            generator.generate(
                num_structures = 5,
                stoichiometry = { 'Mg': num_atoms_Mg, 'O': num_atoms_O },
                method_ratio = {"void": 0.1, "rand": 0.01, "walk": 0.25, "grow": 0.25, "min": 1.0},
                verbose = 0,
            )

            # print the number of structures generated
            print("Total number of structures generated: ", generator.num_structures)
            generated_structures = generator.get_structures(calc)
            num_structures_new = len(generated_structures)

            # check if directory iteration[iter] exists, if not create it
            iterdir = f"iteration{iter}/"
            if not os.path.exists(iterdir):
                os.makedirs(iterdir)
            generator.print_settings(iterdir+"generator_settings.txt")

            # set up list of energies
            energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
            energy_rlxd = np.zeros(num_structures_new - num_structures_old)
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
                unrlxd_structures.append(generated_structures[num_structures_old + i].copy())
                unrlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[num_structures_old + i],
                    energy=generated_structures[num_structures_old + i].get_potential_energy(),
                    forces=generated_structures[num_structures_old + i].get_forces()
                )
            
            # Start parallel execution
            print("Starting parallel execution")
            results = Parallel(n_jobs=5)(
                delayed(process_structure)(i, generated_structures[i].copy(), num_structures_old, calc_params, optimise_structure, iteration=seed)
                for i in range(num_structures_old, num_structures_new)
            )

            # Wait for all futures to complete
            for j, result in enumerate(results):
                generated_structures[num_structures_old + j], energy_unrlxd[j], energy_rlxd[j] = result
                if generated_structures[num_structures_old + j] is None:
                    print("Structure failed the checks")
                    continue
                rlxd_structures.append(generated_structures[num_structures_old + j].copy())
                rlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[j+num_structures_old],
                    energy=generated_structures[num_structures_old + j].get_potential_energy(),
                    forces=generated_structures[num_structures_old + j].get_forces()
                )
            print("All futures completed")

            # Remove structures that failed the checks
            for j, atoms in reversed(list(enumerate(generated_structures))):
                if j < num_structures_old:
                    continue
                if atoms is None:
                    energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                    energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                    del generated_structures[j]
                    del unrlxd_structures[j]
                    del rlxd_structures[j]
                    generator.remove_structure(j)
            num_structures_new = len(generated_structures) 

            # write the structures to files
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
                # append energy per atom to the 'energies_unrlxd_filename' file
                with open(energies_unrlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
                # append energy per atom to the 'energies_rlxd_filename' file
                with open(energies_rlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

            # update the distribution functions
            print("Updating distributions")
            generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)

            # print the new distribution functions to a file
            print("Printing distributions")
            generator.distributions.write_dfs(iterdir+"distributions.txt")
            generator.distributions.write_2body(iterdir+"df2.txt")
            generator.distributions.write_3body(iterdir+"df3.txt")
            generator.distributions.write_4body(iterdir+"df4.txt")
            generator.distributions.deallocate_systems()

            # update the number of structures generated
            num_structures_old = num_structures_new

    # Write the final distribution functions to a file
    generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

    # Read energies from the file
    with open(energies_rlxd_filename, "r") as energy_file:
        energies = energy_file.readlines()

    # Parse and sort the energies
    energies = [line.strip().split() for line in energies]
    energies = sorted(energies, key=lambda x: float(x[1]))

    # Write the sorted energies back to the file
    with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
        for entry in energies:
            energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

    # Write the structures to files
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
    print("All generated and relaxed structures written")

    print("Learning complete")


# This script demonstrates how to use the raffle generator to generate graphite structures

# The script reads a host structure from a POSCAR file, and sets it as the host structure for the generator.

# import standard libraries
import os
import numpy as np

# import raffle library
from raffle.generator import raffle_generator

# import ASE (Atomic Simulation Environment) modules
from ase import Atoms
from ase.io import read, write
from ase.optimize import BFGS

# import CHGNet calculator (for MLP energy evaluation)
from chgnet.model.dynamics import CHGNetCalculator

# load CHGNet calculator
print("Initialising CHGNet calculator")
calculator = CHGNetCalculator(model=None)

print("Initialising raffle generator")
generator = raffle_generator()

print("Reading host")
host = read("hosts/POSCAR_1x5_orthorhombic_11.0A_separation")
host.calc = calculator
print("host energy: ", host.get_potential_energy())

generator.set_host(host)
print("Host read")

print("Setting element energies")
generator.distributions.set_element_energies(
    {
        'C':   -9.0266865, # -1.263,
        'Mg':  -1.5478236, # -0.008,
        'O':   -4.3707458, #-1.888
    }
)

# ## BOND SETTER
# generator.distributions.set_bond_radii(
#     {
#         ('C', 'C'): 1.5,
#         ('C', 'Mg'): 2.0,
#         ('C', 'O'): 1.5,
#         ('Mg', 'Mg'): 2.0,
#         ('Mg', 'O'): 2.0,
#         ('O', 'O'): 1.5
#     }
# )

# set energy scale
generator.distributions.set_kBT(0.5)
# set the distribution function widths (2-body, 3-body, 4-body)
generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

# set the radii tolerances for the 3-body and 4-body distributions
#    these are the lower and upper bounds for the bond radii
#    the first two values are the lower and upper tolerance on covalent bond radii for 3-body distributions
#    the last two values are the lower and upper tolerance on covalent bond radii for 4-body distributions
#    the max bondlength cutoff is an upper limit, so you can turn off 3-body and 4-body distributions by 
#       setting all the values above 100.0 (just to be safe)
generator.distributions.set_radius_distance_tol([1.5, 2.5, 3.0, 6.0])

print("Reading database")
database = read("database.xyz", index=":")
num_database = len(database)

print("Database read")

print("Setting database")
generator.distributions.create(database, deallocate_systems=False)
print("Database set")

print("Printing distributions")
generator.distributions.write("distributions.txt")
generator.distributions.write_2body("df2.txt")
generator.distributions.write_3body("df3.txt")
generator.distributions.write_4body("df4.txt")
# exit(0)

print("Checking element energies")
print(generator.distributions.get_element_energies())

print("Checking bond radii")
print(generator.distributions.get_bond_radii())

print("Setting bins (discretisation of host cell)")
generator.set_grid(grid_spacing=0.2, grid_offset=[0.0, 0.0, 0.0])
print(generator.grid)

print("Setting stoichiometry to insert")
stoich_dict = { 'Mg': 20, 'O': 20 }
generator.distributions.best_energy = 0.0

# generate structures
num_structures_old = 0
optimise_structure = False
for iter in range(10):
    print(f"Iteration {iter}")
    print("Generating...")
    # this is the main function to generate structures
    generator.generate(num_structures=10, stoichiometry=stoich_dict, seed=0+iter, verbose=0, method_ratio={"void":1.0, "walk":1.0, "min":1.0})
    print("Generated")

    print("Getting structures")
    print("number of structures supposed to be generated: ", generator.num_structures)
    generated_structures = generator.get_structures()
    print("actual number allocated: ",len(generated_structures))
    print("Got structures")

    # check if directory iteration[iter] exists, if not create it
    iterdir = f"iteration{iter}/"
    if not os.path.exists(iterdir):
        os.makedirs(iterdir)

    # get energies using MLPs and optimise the structures
    num_structures_new = len(generated_structures)
    for i, atoms in enumerate(generated_structures):
        if(i < num_structures_old):
            continue
        inew = i - num_structures_old
        atoms.calc = calculator
        if optimise_structure:
            optimizer = BFGS(atoms, trajectory = "traje.traj")
            optimizer.run(fmax=0.05)
            print(f"Structure {inew} optimised")
        atoms.get_potential_energy()
        print(f"Structure {inew} energy: {atoms.get_potential_energy()}")
        write(iterdir+f"POSCAR_{inew}", atoms)

    # update the distribution functions
    print("Updating distributions")
    generator.distributions.update(generated_structures[num_structures_old:], deallocate_systems=False)

    # print the new distribution functions to a file
    print("Printing distributions")
    generator.distributions.write(iterdir+"distributions.txt")
    generator.distributions.write_2body(iterdir+"df2.txt")
    generator.distributions.write_3body(iterdir+"df3.txt")
    generator.distributions.write_4body(iterdir+"df4.txt")

    # update the number of structures generated
    num_structures_old = num_structures_new



# from mace.calculators import mace_mp
# from ase.calculators.vasp import Vasp
from chgnet.model import CHGNetCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build, Atoms
from ase.optimize import FIRE
from ase.io import write
from ase.visualize import view
import numpy as np
import os
from concurrent.futures import ProcessPoolExecutor, wait, as_completed
from multiprocessing import Process
from copy import deepcopy
from multiprocessing import Queue
from joblib import Parallel, delayed

import logging
logging.basicConfig(level=logging.DEBUG)

def runInParallel(*fns):
    proc = []
    results = []
    for fn in fns:
        p = Process(target=fn)
        p.start()
        proc.append(p)
    for p in proc:
        results.append(p.join())

    print("All processes finished")
    print(results)


def process_structure(i, atoms, calc_params, optimise_structure, calc):
    # Check if the structure has already been processed
    
    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    atoms.calc = calc
    # positions_initial = atoms.get_positions()

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{i}.traj", logfile=f"optimisation{i}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None
    
    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None
    
    if abs(energy_rlxd - energy_unrlxd) > 5.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None
    
    return atoms, energy_unrlxd, energy_rlxd

# crystal_structures = [
#     'sc', 'fcc', 'bcc', 'hcp',
#     'diamond', 'zincblende', 'rocksalt', 'cesiumchloride',
#     'fluorite', 'wurtzite', 'tetragonal', 'orthorhombic',
#     'bct', 'rhombohedral', 'mcl'
# ]



if __name__ == "__main__":

    calc_params = {}
    calc = CHGNetCalculator()
    # calc_params = {
    #     "model": "large",
    #     "dispersion": False,
    #     "default_dtype": "float32",
    #     "device": 'cpu'
    # }
    # calc = mace_mp(**calc_params)
    # calc_params = {
    #     "command": "$HOME/DVASP/vasp.6.4.3/bin/vasp_std",
    #     # "label": "iteration",
    #     # "txt": "stdout.o",
    #     "xc": 'pbe',
    #     "setups": {'C': ''},
    #     "kpts": (3, 3, 3),
    #     "encut": 400,
    #     "istart": 0,
    #     "icharg": 0,
    # }
    # ## DON'T FORGET TO EXPORT THE VASP_PP_PATH
    # calc = Vasp(**calc_params, label="tmp", directory="tmp", txt="stdout.o")

    #crystal_structures = [
    #    'orthorhombic', 'diamond',
    #    'bct', 'sc',
    #    'fcc', 'bcc', 'hcp',
    #]

    """
    This needs to be changed to set the material cells that will be searched through.
    The crystal_structures list contains the crystal structures that will be used to generate the hosts.
    """
    crystal_structures = [
        'orthorhombic', 'hcp',
    ]

    """
    This needs to be changed to set the lattice constants that will be searched through.
    The lattice_constants list contains the lattice constants that will be used to generate the hosts.
    """
    lattice_constants = np.linspace(3.1, 5.4, num=6)

    """
    This needs to be changed to set the values of the method_ratio.
    """
    #void_val; rand_val; walk_val; grow_val; min_val
    ##############[  V,   R,   W,   G,   M]    
    method_dicts = [
        {"void":  1.0, "rand":  0.0, "walk":  0.0, "grow":  0.0, "min":  0.0},
        {"void":  0.0, "rand":  1.0, "walk":  0.0, "grow":  0.0, "min":  0.0},
        {"void":  0.0, "rand":  0.0, "walk":  1.0, "grow":  0.0, "min":  0.0},
        {"void":  0.0, "rand":  0.0, "walk":  0.0, "grow":  1.0, "min":  0.0},
        {"void":  0.0, "rand":  0.0, "walk":  0.0, "grow":  0.0, "min":  1.0},
        {"void": 85.0, "rand": 15.0, "walk":  0.0, "grow":  0.0, "min":  0.0},
        {"void":  0.0, "rand": 15.0, "walk": 85.0, "grow":  0.0, "min":  0.0},
        {"void":  0.0, "rand": 15.0, "walk":  0.0, "grow": 85.0, "min":  0.0},
        {"void":  0.0, "rand": 15.0, "walk":  0.0, "grow":  0.0, "min": 85.0},
    ]

    hosts = []
    for crystal_structure in crystal_structures:
        print(f'Crystal structure: {crystal_structure}')
        for a in lattice_constants:
            b = a
            c = a
            atom = build.bulk(
                    name = 'Al',
                    crystalstructure = crystal_structure,
                    a = a,
                    b = b,
                    c = c,
            )
            hosts.append(Atoms('Al', positions=[(0, 0, 0)], cell=atom.get_cell(), pbc=True, calculator=calc))
            # hosts[-1].set_pbc(True)
            # hosts[-1].calc = calc
            print(hosts[-1])

    print("number of hosts: ", len(hosts))

    optimise_structure = True
    mass = 26.9815385
    density = 1.61 # u/A^3
    # num_atoms = 7

    # get the current directory
    cwd = os.getcwd()
    print(f"Current working directory: {cwd}")

    seed = 0
    for methods in method_dicts:
        # move to the cwd
        os.chdir(cwd)
        print(f"Changed directory to {cwd}")

        # print the methods
        print(f"Method: {methods}")

        # make method directory
        method_dir = f"method_v{methods['void']}_r{methods['rand']}_w{methods['walk']}_g{methods['grow']}_m{methods['min']}/"
        if not os.path.exists(method_dir):
            os.makedirs(method_dir)
        os.chdir(method_dir)
        print(f"Changed directory to {method_dir}")
        
        energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
        energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
        generator = raffle_generator()
        generator.distributions.set_element_energies(
            {
                'Al': 0.0
            }
        )
        # set energy scale
        generator.distributions.set_kBT(0.4)
        # set the distribution function widths (2-body, 3-body, 4-body)
        generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

        initial_database = [Atoms('Al', positions=[(0, 0, 0)], cell=[8, 8, 8], pbc=True)]
        initial_database[0].calc = calc

        generator.distributions.create(initial_database)

        if os.path.exists(energies_rlxd_filename):
            with open(energies_rlxd_filename, "w") as energy_file:
                pass
        else:
            open(energies_rlxd_filename, "w").close()

        if os.path.exists(energies_unrlxd_filename):
            with open(energies_unrlxd_filename, "w") as energy_file:
                pass
        else:
            open(energies_unrlxd_filename, "w").close()

        # delete .traj files if they exist
        if os.path.exists(f"unrlxd_structures_seed{seed}.xyz"):
            os.remove(f"unrlxd_structures_seed{seed}.xyz")
        if os.path.exists(f"rlxd_structures_seed{seed}.xyz"):
            os.remove(f"rlxd_structures_seed{seed}.xyz")

        num_structures_tot = 0
        unrlxd_structures = []
        rlxd_structures = []
        iter2 = -1
        generator.init_seed(put=seed)
        for iter in range(10):
            for host in hosts:
                generator.set_host(host)
                volume = host.get_volume()

                num_atoms = round(density * volume / mass) - 1
                if(num_atoms < 1):
                    continue
                iter2 += 1
                print(f"Volume: {volume}")
                print(f"Number of atoms: {num_atoms}")
    
                generated_structures = generator.generate(
                    num_structures = 5,
                    stoichiometry = { 'Al': num_atoms },
                    method_ratio = methods,
                    verbose = 0,
                    calc = calc,
                )

                # print the number of structures generated
                print("Total number of structures generated: ", generator.num_structures)
                num_structures_new = len(generated_structures)

                # check if directory iteration[iter] exists, if not create it
                iterdir = f"iteration{iter2}/"
                if not os.path.exists(iterdir):
                    os.makedirs(iterdir)
                generator.print_settings(iterdir+"generator_settings.txt")

                # set up list of energies
                for i in reversed(range(len(generated_structures))):
                    if generated_structures[i] is None:
                        print(f"Structure {i} is None, skipping")
                        del generated_structures[i]
                        continue
                    write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[i])
                    print(f"Structure {i} energy per atom: {generated_structures[i].get_potential_energy() / len(generated_structures[i])}")

            
                # Start parallel execution
                print("Starting parallel execution")
                results = Parallel(n_jobs=5)(
                    delayed(process_structure)(i, generated_structures[i].copy(), calc_params, optimise_structure, calc=calc)
                    for i in range(len(generated_structures))
                )

                # Wait for all futures to complete
                energies_unrlxd = []
                energies_rlxd = []
                for j, result in enumerate(results):
                    rlxd_generated_structure, energy_unrlxd, energy_rlxd = result
                    if rlxd_generated_structure is None:
                        print("Structure failed the checks")
                        continue
                    else:
                        unrlxd_structures.append(generated_structures[j].copy())
                        unrlxd_structures[-1].calc = SinglePointCalculator(
                            generated_structures[j],
                            energy=generated_structures[j].get_potential_energy(),
                            forces=generated_structures[j].get_forces()
                        )
                        rlxd_structures.append(rlxd_generated_structure.copy())
                        rlxd_structures[-1].calc = SinglePointCalculator(
                            rlxd_generated_structure,
                            energy=rlxd_generated_structure.get_potential_energy(),
                            forces=rlxd_generated_structure.get_forces()
                        )
                        energies_unrlxd.append(energy_unrlxd)
                        energies_rlxd.append(energy_rlxd)
                print("All futures completed")
                num_structures_new = len(energies_unrlxd)


                # write the structures to files
                for i in range(num_structures_new):
                    write(iterdir+f"POSCAR_{i}", rlxd_structures[num_structures_tot + i])
                    print(f"Structure {i} energy per atom: {energies_rlxd[i]}")
                    # append energy per atom to the 'energies_unrlxd_filename' file
                    with open(energies_unrlxd_filename, "a") as energy_file:
                        energy_file.write(f"{i+num_structures_tot} {energies_unrlxd[i]}\n")
                    # append energy per atom to the 'energies_rlxd_filename' file
                    with open(energies_rlxd_filename, "a") as energy_file:
                        energy_file.write(f"{i+num_structures_tot} {energies_rlxd[i]}\n")

                # update the distribution functions
                print("Updating distributions")
                generator.distributions.update(rlxd_structures[num_structures_tot:], from_host=False, deallocate_systems=False)

                # print the new distribution functions to a file
                print("Printing distributions")
                generator.distributions.write_dfs(iterdir+"distributions.txt")
                generator.distributions.write_2body(iterdir+"df2.txt")
                generator.distributions.write_3body(iterdir+"df3.txt")
                generator.distributions.write_4body(iterdir+"df4.txt")
                generator.distributions.deallocate_systems()

                write(f"unrlxd_structures_seed{seed}.xyz", unrlxd_structures[num_structures_tot:], append=True)
                write(f"rlxd_structures_seed{seed}.xyz", rlxd_structures[num_structures_tot:], append=True)

                # update the number of structures generated
                num_structures_tot += num_structures_new


        generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

        # Read energies from the file
        with open(energies_rlxd_filename, "r") as energy_file:
            energies = energy_file.readlines()

        # Parse and sort the energies
        energies = [line.strip().split() for line in energies]
        energies = sorted(energies, key=lambda x: float(x[1]))

        # Write the sorted energies back to the file
        with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
            for entry in energies:
                energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

        write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
        write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
        print("All generated and relaxed structures written")

    print("Learning complete")


# https://agox.gitlab.io/agox/command_line_tools/command_line.html

import matplotlib

matplotlib.use("Agg")


import numpy as np
from ase import Atoms
from ase.io import write

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator
from agox.postprocessors import MinimumDistPostProcess


database_index = 0
iteration_directory = "iteration"

##############################################################################
# Calculator
##############################################################################

# from mace.calculators import mace_mp
from chgnet.model import CHGNetCalculator
from ase.build import bulk
import os

# calc = mace_mp()
calc = CHGNetCalculator()

##############################################################################
# System & general settings:
##############################################################################


post1 = MinimumDistPostProcess(order=1.5, c1=0.75)
post2 = MinimumDistPostProcess(order=2.5, c1=0.75)

a = 3.19
MoS2 = Atoms('Mo', positions=[(0, 0, 0)], cell=[
    [a * 0.5, a * -np.sqrt(3.0)/2.0, 0.0],
    [a * 0.5, a * np.sqrt(3.0)/2.0, 0.0],
    [0.0, 0.0, 13.1]
    ], pbc=True, calculator=calc)

# Make an empty template with a cubic cell of size 6x6x6 with periodic boundary conditions
# in all directions
iteration = 0
template = Atoms("",
                 cell=[
                    [a * 0.5, a * -np.sqrt(3.0)/2.0, 0.0],
                    [a * 0.5, a * np.sqrt(3.0)/2.0, 0.0],
                    [0.0, 0.0, 13.1]
                 ],
                 pbc=True)


# Confinement cell matches the template cell in all dimensions and is placed at
# the origin of the template cell.
confinement_cell = template.cell.copy()
confinement_corner = np.array([0, 0, 0])

environment = Environment(
    template=template,
    symbols="Mo2S4",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[True, True, True],  # Confinement is periodic in all directions.
)

# Database
db_directory = iteration_directory+"{}".format(iteration)+"/"
if not os.path.exists(db_directory):
    os.makedirs(db_directory)
db_path = db_directory+"db{}.db".format(database_index)  # From input argument!

for seed in range(1):
    database = Database(filename=db_path, order=3, initialize=True)

    ##############################################################################
    # Search Settings:
    ##############################################################################

    random_generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

    # Wont relax fully with steps:5 - more realistic setting would be 100+.
    evaluator = LocalOptimizationEvaluator(
        calc,
        gets={"get_key": "candidates"},
        optimizer_run_kwargs={"fmax": 0.05, "steps": 100},
        store_trajectory=True,
        order=2,
        constraints=environment.get_constraints(),
    )

    ##############################################################################
    # Let get the show running!
    ##############################################################################

    agox = AGOX(random_generator, database, evaluator, seed=seed)

    agox.run(N_iterations=1000)

    database.restore_to_memory()
    structure_data = database.get_all_structures_data()
    # check iteration number has changed
    iteration_check = -1
    unrlxd_structures = []
    rlxd_structures = []
    for idx, structure in enumerate(structure_data):
        if structure["iteration"] != iteration_check:
            iteration_check = structure["iteration"]
            unrlxd_structures.append(database.db_to_atoms(structure))
            if idx != 0:
                rlxd_structures.append(database.db_to_atoms(structure_data[idx-1]))
    if len(rlxd_structures) != len(unrlxd_structures):
        rlxd_structures.append(database.db_to_atoms(structure_data[-1]))

    for structure in unrlxd_structures:
        structure.calc = None
    for structure in rlxd_structures:
        structure.calc = None
    print("Unrelaxed structures", len(unrlxd_structures))
    print("Relaxed structures", len(rlxd_structures))
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)

# database.restore_to_memory()
# structures = database.get_all_candidates() # this is an Atoms object with some more stuff



# This script demonstrates how to use the raffle generator to generate BaTiO3 perovskite structures

# The script reads a host structure from a POSCAR file, and sets it as the host structure for the generator.

# import standard libraries
import os
import numpy as np

# import raffle library
from raffle.generator import raffle_generator

# import ASE (Atomic Simulation Environment) modules
from ase import Atoms
from ase.io import read, write
from ase.optimize import BFGS

# import CHGNet calculator (for MLP energy evaluation)
from chgnet.model.dynamics import CHGNetCalculator

# load CHGNet calculator
print("Initialising CHGNet calculator")
calculator = CHGNetCalculator(model=None)

# set up an instance of the raffle generator
print("Initialising raffle generator")
generator = raffle_generator()

# read the host structure from a POSCAR file
print("Reading host")
host = read("POSCAR_host")
host.calc = calculator
print("host energy: ", host.get_potential_energy())

generator.set_host(host)
print("Host read")

# generate bulk BaTiO3 and get its energy
reference_bulk = Atoms("BaTiO3", positions=[[0.0, 0.0, 0.0], 
                                        [2.005, 2.005, 2.005],
                                        [2.005, 2.005, 0.000],
                                        [2.005, 0.000, 2.005],
                                        [0.000, 2.005, 2.005]
                                    ], cell=[4.01, 4.01, 4.01], pbc=True)
reference_bulk.calc = calculator

# generate the bulks of the individual elements and get their energies
Ba_reference = Atoms('Ba4', positions=[
    [1.2746913321062139, 0.0, 1.797511295],
    [3.7591171278937865, 0.0, 5.392533885000001],
    [3.7915955621062145, 3.59177373, 1.7975112950000005],
    [1.242212897893787, 3.59177373, 5.392533885000001]
    ], cell=[5.03380846, 7.18354746, 7.19004518], pbc=True)
Ba_reference.calc = calculator
Ba_reference_energy = Ba_reference.get_potential_energy() / len(Ba_reference)
Ti_reference = Atoms("Ti3", positions=[
    [0.0, 0.0, 0.0],
    [2.2836874152833575, 1.3184875439588069, 1.413122135],
    [2.283687415283358, -1.3184875439588073, 1.413122135]
    ], cell=[[2.2836874152833575, -3.955462631876421, 0.0], [2.2836874152833575, 3.955462631876421, 0.0], [0.0, 0.0, 2.82624427]], pbc=True)
Ti_reference.calc = calculator
Ti_reference_energy = Ti_reference.get_potential_energy() / len(Ti_reference)
O_reference = Atoms("O2", positions=[
    [0.0, 0.0, 0.0],
    [1.23, 0.0, 0.0]], cell=[10, 10, 10], pbc=False)
O_reference.calc = calculator
O_reference_energy = O_reference.get_potential_energy() / len(O_reference)

# print energies
print(f"Ba_reference_energy: {Ba_reference_energy}")
print(f"Ti_reference_energy: {Ti_reference_energy}")
print(f"O_reference_energy: {O_reference_energy}")
print(f"BaTiO3_reference_energy: {reference_bulk.get_potential_energy()}")

# set the element reference energies
print("Setting element energies")
generator.distributions.set_element_energies(
    {
        'Ba': Ba_reference_energy,
        'Ti': Ti_reference_energy,
        'O': O_reference_energy
    }
)

# set energy scale
generator.distributions.set_kBT(0.2)
# set the distribution function widths (2-body, 3-body, 4-body)
generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

# set the radii tolerances for the 3-body and 4-body distributions
#    these are the lower and upper bounds for the bond radii
#    the first two values are the lower and upper tolerance on covalent bond radii for 3-body distributions
#    the last two values are the lower and upper tolerance on covalent bond radii for 4-body distributions
#    the max bondlength cutoff is an upper limit, so you can turn off 3-body and 4-body distributions by 
#       setting all the values above 100.0 (just to be safe)
generator.distributions.set_radius_distance_tol([1.5, 2.5, 3.0, 6.0])

# set the distribution function widths (2-body, 3-body, 4-body)
print("Reading database")
use_database = False
if use_database:
    database = read("database.xyz", index=":")
    for i, atoms in enumerate(database):
        # reset energy to use CHGNet
        atoms.calc = calculator
else:
    database = []
    database.append(reference_bulk)
print("Database read")

# create the distribution functions
print("Setting database")
generator.distributions.create(database, deallocate_systems=False)
print("Database set")

# print the distribution functions to a file
print("Printing distributions")
generator.distributions.write("distributions.txt")
generator.distributions.write_2body("df2.txt")
generator.distributions.write_3body("df3.txt")
generator.distributions.write_4body("df4.txt")

# check the element energies and bond radii
print("Checking element energies")
print(generator.distributions.get_element_energies())
print("Checking bond radii")
print(generator.distributions.get_bond_radii())

# set the grid for the host cell
print("Setting bins (discretisation of host cell)")
generator.set_grid(grid_spacing=0.1, grid_offset=[0.0, 0.0, 0.0])
print(generator.grid)

# set the stoichiometry for the structures to be generated
print("Setting stoichiometry to insert")
stoich_dict = { 'Ba': 1, 'Ti': 1, 'O': 3 }

# generate structures
num_structures_old = 0
optimise_structure = False
for iter in range(1):
    print(f"Iteration {iter}")
    print("Generating...")
    # this is the main function to generate structures
    generator.generate(num_structures=1, stoichiometry=stoich_dict, seed=0+iter, verbose=1, method_ratio={"void":0.001, "walk":0.0, "min":1.0})
    print("Generated")

    print("Getting structures")
    print("number of structures supposed to be generated: ", generator.num_structures)
    generated_structures = generator.get_structures()
    print("actual number allocated: ",len(generated_structures))
    print("Got structures")

    # check if directory iteration[iter] exists, if not create it
    iterdir = f"iteration{iter}/"
    if not os.path.exists(iterdir):
        os.makedirs(iterdir)

    # get energies using MLPs and optimise the structures
    num_structures_new = len(generated_structures)
    for i, atoms in enumerate(generated_structures):
        if(i < num_structures_old):
            continue
        inew = i - num_structures_old
        atoms.calc = calculator
        if optimise_structure:
            optimizer = BFGS(atoms, trajectory = "traje.traj")
            optimizer.run(fmax=0.05)
            print(f"Structure {inew} optimised")
        atoms.get_potential_energy()
        print(f"Structure {inew} energy: {atoms.get_potential_energy()}")
        write(iterdir+f"POSCAR_{inew}", atoms)

    # update the distribution functions
    print("Updating distributions")
    generator.distributions.update(generated_structures[num_structures_old:], deallocate_systems=False)

    # print the new distribution functions to a file
    print("Printing distributions")
    generator.distributions.write(iterdir+"distributions.txt")
    generator.distributions.write_2body(iterdir+"df2.txt")
    generator.distributions.write_3body(iterdir+"df3.txt")
    generator.distributions.write_4body(iterdir+"df4.txt")

    # update the number of structures generated
    num_structures_old = num_structures_new



import matplotlib

matplotlib.use("Agg")

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
import shephex
from ase import build

@shephex.chain()
def raffle_agox(index=42, void=0.1, rand=0.01, walk=0.25, grow=0.25, min=1.0):


    # Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
    # from argparse import ArgumentParser
    # parser = ArgumentParser()
    # parser.add_argument('-i', '--run_idx', type=int, default=0)
    # args = parser.parse_args()

    ##############################################################################
    # System & generator settings:
    ##############################################################################

    import numpy as np
    from agox.generators.raffle import RaffleGenerator, bounds_to_confinement
    # from mace.calculators import mace_mp
    from agox.utils.replica_exchange.priors import get_prior
    from ase.io import read

    ## Set seed and database-index
    seed = index
    database_index = index
    np.random.seed(seed)

    ## Set up the MACE calculator
    mace = get_prior("mace-mpa")
    # calc_params = { 'model':  '/FULL/PATH/TO/MODEL/FILE' }
    # mace = mace_mp(**calc_params)

    ## Set up the method ratio dictionary and generate a unique ID
    method_ratio = {
        "void": void,
        "rand": rand,
        "walk": walk,
        "grow": grow,
        "min": min
    }
    # generate unique id from placement dict
    method_id = "_".join(f"{k}{v:g}" for k, v in sorted(method_ratio.items()))
    print("method ratio:", method_ratio)
    print("method_id:", method_id)

    ## Set up the host structure (CHANGE THIS TO YOUR HOST STRUCTURE LOCATION)
    template = read("${HOME}/DAGOX/Dgraphene_grain_boundary/POSCAR_host_gb")
    template.calc = mace
    template.set_pbc(True)

    graphene = read("${HOME}/DAGOX/Dgraphene_grain_boundary/POSCAR_graphene")

    ## Set up the reference energies for C and H
    h2 = build.molecule("H2")
    graphene.calc = mace
    C_reference_energy = graphene.get_potential_energy() / len(graphene)
    h2.calc = mace
    H_reference_energy = h2.get_potential_energy() / len(h2)
    element_energies = {
        'C': C_reference_energy,
        'H': H_reference_energy,
    }
    print("element reference energies:", element_energies)

    ## Set up the species to place in the host structure
    species_to_place={'C' : 15}
    symbols = ""
    for key in species_to_place.keys():
        symbols += key
        symbols += str(species_to_place[key])

    ## Set up the bounding box for the confinement
    confinement_corner, confinement_cell = bounds_to_confinement([[0.2, 0, 0.18], [0.3, 1, 0.33]],template)

    ## Set up the environment
    environment = Environment(
        template=template,
        symbols=symbols,
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
        fix_template = False,
    )

    # Database
    db_path = f"../db{database_index}_{method_id}.db"  # From input argument!
    database = Database(filename=db_path, order=5)
    database.restore_to_memory()

    ## Set up the RaffleGenerator
    generator = RaffleGenerator(
        **environment.get_confinement(),
        host=template,
        database=database,
        environment=environment,
        element_energies=element_energies,
        transfer_data=[graphene],
        order=1,
        width=[0.04, np.pi/160.0, np.pi/160.0],
        kBT=0.2,
        n_structures=5,
        from_host=False,
        sampler=None,
        train=True,
        seed=seed,
        method_ratio = method_ratio
    )

    ##############################################################################
    # Search Settings:
    ##############################################################################

    ## Set up the structural optimisation evaluator (defaults to ASE's BFGS optimizer)
    # from ase.optimize import FIRE # Uncomment to use FIRE instead of BFGS
    evaluator = LocalOptimizationEvaluator(
        mace,
        # optimizer = FIRE, # Uncomment to use FIRE instead of BFGS
        gets={"get_key": "candidates"},
        store_trajectory=False,
        optimizer_run_kwargs={"fmax": 0.05, "steps": 200},
        order=3,
        number_to_evaluate=5,
        constraints=environment.get_constraints(),
        fix_template = False,
    )

    ## Set up post-processors to filter the generated structures.
    # This checks the minimum distance between atoms in the generated structures.
    from agox.postprocessors.minimum_dist import MinimumDistPostProcess
    c1 = 0.6
    c2 = 5
    minimum_dist = MinimumDistPostProcess(
        c1=c1,
        c2=c2,
        order=1.5,
    )
    minimum_dist2 = MinimumDistPostProcess(
        c1=c1,
        gets = {"get_key" : "evaluated_candidates"},
        sets = {"set_key" : "evaluated_candidates"},
        c2=c2,
        order=3.5,
    )

    ##############################################################################
    # Let get the show running!
    ##############################################################################

    ## Set up the AGOX search workflow
    agox = AGOX(generator, minimum_dist, minimum_dist2, database, evaluator, seed=seed)

    ## Run the AGOX search for N_iterations
    agox.run(N_iterations=40)



if __name__ == "__main__":

    walkers = 16

    ## Set up the experiments to run
    # This is an example of how to set up a series of experiments with different
    # method ratios and different seeds.
    experiments = (raffle_agox(directory='method_ratio/').
                   zip(void=[1.0, 0.1,  0.0, 0.1]).
                   zip(rand=[0.1, 0.01, 1.0, 0.0]).
                   zip(walk=[0.5, 0.25, 0.0, 0.5]).
                   zip(grow=[0.5, 0.25, 0.0, 0.5]).
                   zip(min =[1.0, 1.0,  0.0, 1.0]).
                   permute(index=range(10))
                   )

    ## Uncomment for Slurm execution
    executor = shephex.executor.SlurmExecutor.from_config(
        "basic.json",
        safety_check=False,
        mem_per_cpu='1024',
        time="1-00:00:00",
        ntasks_per_node=walkers
    )

    # ## Uncomment for local execution
    # executor = shephex.executor.LocalExecutor()

    # Submit the experiments to the executor
    executor.execute(experiments)



# %%
import os
from pathlib import Path
from ase.io import read, write
from ase.visualize import view
from ase.optimize import FIRE
from ase.calculators.singlepoint import SinglePointCalculator
from mace.calculators import mace_mp
from artemis.generator import artemis_generator

script_dir = Path(__file__).resolve().parent

# %%
abrupt = read(script_dir / 'SiGe_abrupt_interface_rescaled.vasp')
abrupt.set_pbc(True)

# %%
generator = artemis_generator()

# %%
loc, axis = generator.get_interface_location(abrupt)

# %%
generator.set_shift_method(
    num_shifts = 20
)
# %%
structures = generator.regenerate(abrupt, verbose=1)

# %%
# make directory for the structures
output_dir = script_dir / "SiGe_shifted_structures"
output_dir.mkdir(parents=True, exist_ok=True)
for i, structure in enumerate(structures):
    write(output_dir / f'SiGe_shifted_{i}.vasp', structure, format='vasp', sort=True, direct=True)

exit()
# %%

# check if mace file exists
if not os.path.exists(script_dir  / ".." / "mace-mpa-0-medium.model"):
    print("MACE-MPA-0 model file not found. Please download the model from the MACE website.")
    print("https://github.com/ACEsuit/mace-foundations/releases/tag/mace_mpa_0")
    exit(1)

# set up the calculator
calc_params = { 'model':  script_dir / ".." / "mace-mpa-0-medium.model" }
calc = mace_mp(**calc_params)

# %%
structures = [ abrupt.copy() ]
structures.extend(generator.get_structures(calculator = calc))

for structure in structures:
    structure.calc = calc
    structure.calc = SinglePointCalculator(
        structure,
        energy=structure.get_potential_energy(),
        forces=structure.get_forces()
    )


write('SiGe_shifted_unrlxd.traj', structures)

view(structures)

# %%

for structure in structures:
    structure.calc = calc
    optimizer = FIRE(structure)
    optimizer.run(fmax=0.05, steps=200)
    structure.calc = SinglePointCalculator(
        structure,
        energy=structure.get_potential_energy(),
        forces=structure.get_forces()
    )

write('SiGe_shifted_rlxd.traj', structures)



from mace.calculators import mace_mp
from chgnet.model import CHGNetCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build
from ase.optimize import FIRE
from ase.io import read, write
import numpy as np
import os
from joblib import Parallel, delayed
from ase.constraints import FixAtoms
from pathlib import Path

script_dir = Path(__file__).resolve().parent

import logging
logging.basicConfig(level=logging.DEBUG)

# Function to relax a structure
def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration, calc):
    # Check if the structure has already been processed
    if i < num_structures_old:
        return None, None, None

    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None

    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None

    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None

    return atoms, energy_unrlxd, energy_rlxd


if __name__ == "__main__":

    # check if mace file exists
    if not os.path.exists(script_dir  / ".." / ".." / "mace-mpa-0-medium.model"):
        print("MACE-MPA-0 model file not found. Please download the model from the MACE website.")
        print("https://github.com/ACEsuit/mace-foundations/releases/tag/mace_mpa_0")
        exit(1)

    # set up the calculator
    # calc_params = {}
    # calc = CHGNetCalculator()
    calc_params = { 'model':  script_dir/ ".." / ".." / "mace-mpa-0-medium.model" }
    calc = mace_mp(**calc_params)

    # set up the hosts
    hosts = []
    host = read(script_dir  / ".." / "POSCAR_host_gb")
    hosts.append(host)

    # set the parameters for the generator
    generator = raffle_generator()
    generator.distributions.set_history_len(10)
    graphene = read(script_dir / ".." / "POSCAR_graphene")
    h2 = build.molecule("H2")
    graphene.calc = calc
    C_reference_energy = graphene.get_potential_energy() / len(graphene)
    h2.calc = calc
    H_reference_energy = h2.get_potential_energy() / len(h2)
    generator.distributions.set_element_energies(
        {
            'C': C_reference_energy,
            'H': H_reference_energy,
        }
    )

    # set energy scale
    generator.distributions.set_kBT(0.2)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width([0.04, np.pi/160.0, np.pi/160.0])

    # set the initial database
    initial_database = [graphene]
    generator.distributions.create(initial_database, deallocate_systems=False)

    # set the parameters for the generator
    seed = 0

    # check if the energies file exists, if not create it
    energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
    energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"

    if os.path.exists(energies_rlxd_filename):
        with open(energies_rlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_rlxd_filename, "w").close()

    if os.path.exists(energies_unrlxd_filename):
        with open(energies_unrlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_unrlxd_filename, "w").close()

    # initialise the number of structures generated
    iter = -1
    unrlxd_structures = []
    rlxd_structures = []
    num_structures_old = 0
    optimise_structure = True
    converged = False
    generator.init_seed(put=seed)
    # start the iterations, loop over the hosts, then repeat the process X times
    for ival in range(40):
        if converged:
            print("Converged")
            break
        for host in hosts:
            print("setting host")
            generator.set_host(host)
            generator.set_bounds([[0.20, 0, 0.18], [0.30, 1, 0.33]])

            iter += 1
            print(f"Iteration {iter}")

            # generate the structures
            generator.generate(
                num_structures = 5,
                stoichiometry = { 'C': 15 },
                method_ratio = {"void": 0.1, "rand": 0.01, "walk": 0.25, "grow": 0.25, "min": 1.0},
                verbose = 0,
            )

            # print the number of structures generated
            print("Total number of structures generated: ", generator.num_structures)
            generated_structures = generator.get_structures(calc)
            num_structures_new = len(generated_structures)

            # check if directory iteration[iter] exists, if not create it
            iterdir = f"iteration{iter}/"
            if not os.path.exists(iterdir):
                os.makedirs(iterdir)
            generator.print_settings(iterdir+"generator_settings.txt")

            # set up list of energies
            energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
            energy_rlxd = np.zeros(num_structures_new - num_structures_old)
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
                # print(f"Structure {i} energy per unit area: {( generated_structures[num_structures_old + i].get_potential_energy() - Si_slab.get_potential_energy() - Ge_slab.get_potential_energy() ) / (2 * ( cell[0, 0] * cell[1, 1] ) )}")
                # fix all hydrogen atoms to stop them from moving
                c = FixAtoms(indices=[atom.index for atom in generated_structures[num_structures_old + i] if atom.symbol == 'H'])
                generated_structures[num_structures_old + i].set_constraint(c)
                unrlxd_structures.append(generated_structures[num_structures_old + i].copy())
                unrlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[num_structures_old + i],
                    energy=generated_structures[num_structures_old + i].get_potential_energy(),
                    forces=generated_structures[num_structures_old + i].get_forces()
                )

            # Start parallel execution
            print("Starting parallel execution")
            results = Parallel(n_jobs=5)(
                delayed(process_structure)(i, generated_structures[i].copy(), num_structures_old, calc_params, optimise_structure, iteration=seed, calc=calc)
                for i in range(num_structures_old, num_structures_new)
            )

            # Wait for all futures to complete
            for j, result in enumerate(results):
                generated_structures[num_structures_old + j], energy_unrlxd[j], energy_rlxd[j] = result
                if generated_structures[num_structures_old + j] is None:
                    print("Structure failed the checks")
                    continue
                rlxd_structures.append(generated_structures[num_structures_old + j].copy())
                rlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[j+num_structures_old],
                    energy=generated_structures[num_structures_old + j].get_potential_energy(),
                    forces=generated_structures[num_structures_old + j].get_forces()
                )
            print("All futures completed")

            # Remove structures that failed the checks
            for j, atoms in reversed(list(enumerate(generated_structures))):
                if j < num_structures_old:
                    continue
                if atoms is None:
                    energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                    energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                    del generated_structures[j]
                    del unrlxd_structures[j]
                    del rlxd_structures[j]
                    generator.remove_structure(j)
            num_structures_new = len(generated_structures)

            # write the structures to files
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
                # append energy per atom to the 'energies_unrlxd_filename' file
                with open(energies_unrlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
                # append energy per atom to the 'energies_rlxd_filename' file
                with open(energies_rlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

            # update the distribution functions
            print("Updating distributions")
            generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)
            if generator.distributions.is_converged(threshold=1e-3):
                converged = True
                break
            else:
              print("deltas", generator.distributions.history_deltas)

            # print the new distribution functions to a file
            print("Printing distributions")
            generator.distributions.write_dfs(iterdir+"distributions.txt")
            generator.distributions.write_2body(iterdir+"df2.txt")
            generator.distributions.write_3body(iterdir+"df3.txt")
            generator.distributions.write_4body(iterdir+"df4.txt")
            generator.distributions.deallocate_systems()

            # update the number of structures generated
            num_structures_old = num_structures_new


    # Write the final distribution functions to a file
    generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

    # Read energies from the file
    with open(energies_rlxd_filename, "r") as energy_file:
        energies = energy_file.readlines()

    # Parse and sort the energies
    energies = [line.strip().split() for line in energies]
    energies = sorted(energies, key=lambda x: float(x[1]))

    # Write the sorted energies back to the file
    with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
        for entry in energies:
            energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

    # Write the structures to files
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
    print("All generated and relaxed structures written")

    print("Learning complete")



import os
import numpy as np

## import ASE (Atomic Simulation Environment) modules
from ase import Atoms
from ase.io import read, write
from ase.build import bulk


## load calculator
from chgnet.model.dynamics import CHGNetCalculator
print("Initialising CHGNet calculator")
chgnet = CHGNetCalculator()
from mace.calculators import mace_mp
print("Initialising MACE calculator")
calc_params = { 'model': "../example/python_pkg/mace-mpa-0-medium.model" }
mace = mace_mp(**calc_params)
# calc = mace_mp(model="medium", dispersion=False, default_dtype="float32", device='cpu')

## Read the database
print("Reading database")
database = read("../example/data/Si-Ge.xyz", index=":")
database.extend(read("../example/data/Si-Ge_interfaces_vasp.xyz", index=":"))

Si_energy_dft = -10.644538 / 2 # Si POT
Ge_energy_dft = -8.7301811 / 2 # Ge POT

Si_bulk = bulk("Si")
Si_bulk.calc = chgnet
Si_energy_chgnet = Si_bulk.get_potential_energy() / len(Si_bulk)
Ge_bulk = bulk("Ge")
Ge_bulk.calc = chgnet
Ge_energy_chgnet = Ge_bulk.get_potential_energy() / len(Ge_bulk)

Si_bulk = bulk("Si")
Si_bulk.calc = mace
Si_energy_mace = Si_bulk.get_potential_energy() / len(Si_bulk)
Ge_bulk = bulk("Ge")
Ge_bulk.calc = mace
Ge_energy_mace = Ge_bulk.get_potential_energy() / len(Ge_bulk)

Si_energy_GPAW = -5.395989432338489
Ge_energy_GPAW = -4.499373544346001

## Calculate the energies
energies_dft = []
energies_chgnet = []
energies_mace = []
formation_energies_dft = []
formation_energies_chgnet = []
formation_energies_mace = []
for i, atoms in enumerate(database):
    if atoms.calc is None:
        database.remove(atoms)
        continue
    # get number of Si atoms in the structure
    n_Si = atoms.get_atomic_numbers().tolist().count(14)
    n_Ge = atoms.get_atomic_numbers().tolist().count(32)
    if(n_Si == 0 and n_Ge == 0 or n_Si + n_Ge != len(atoms)):
        print(f"Skipping structure {i} as it has no Si or Ge atoms")
        continue
    # if(n_Si == 0 or n_Ge == 0):
    #     print(f"Skipping structure {i} as it has only one type of atom")
    #     continue
    formation_energy_dft = ( atoms.get_potential_energy() - n_Si * Si_energy_dft - n_Ge * Ge_energy_dft ) / len(atoms)
    formation_energies_dft.append(formation_energy_dft)
    energies_dft.append(atoms.get_potential_energy()/len(atoms))
    atoms.calc = chgnet
    formation_energy_chgnet = ( atoms.get_potential_energy() - n_Si * Si_energy_chgnet - n_Ge * Ge_energy_chgnet ) / len(atoms)
    formation_energies_chgnet.append(formation_energy_chgnet)
    atoms.calc = mace
    energies_chgnet.append(atoms.get_potential_energy()/len(atoms))
    formation_energy_mace = ( atoms.get_potential_energy() - n_Si * Si_energy_mace - n_Ge * Ge_energy_mace ) / len(atoms)
    energies_mace.append(atoms.get_potential_energy()/len(atoms))
    formation_energies_mace.append(formation_energy_mace)
    print(n_Si, n_Ge, formation_energy_dft, formation_energy_chgnet, formation_energy_mace)
    # if energies_mlp[-1] - energies_dft[-1] > 3e-1 and energies_mlp[-1] < -7.9:
    #     print(f"Energy difference for structure {i} is {energies_mlp[-1] - energies_dft[-1]}, energy_mace: {energies_mlp[-1]}")
    #     view(atoms)

# database_gpaw = read("../example/data/Si-Ge_interfaces_gpaw.xyz", index=":")
# energies_gpaw = []
# energies_mlp_gpaw = []
# formation_energies_gpaw = []
# formation_energies_mlp_gpaw = []
# for i, atoms in enumerate(database_gpaw):
#     if atoms.calc is None:
#         database_gpaw.remove(atoms)
#         continue
#     # get number of Si atoms in the structure
#     n_Si = atoms.get_atomic_numbers().tolist().count(14)
#     n_Ge = atoms.get_atomic_numbers().tolist().count(32)
#     formation_energy_gpaw = ( atoms.get_potential_energy() - n_Si * Si_energy_GPAW - n_Ge * Ge_energy_GPAW ) / len(atoms)
#     formation_energies_gpaw.append(formation_energy_gpaw)
#     energies_gpaw.append(atoms.get_potential_energy()/len(atoms))
#     atoms.calc = calc
#     formation_energy_mlp_gpaw = ( atoms.get_potential_energy() - n_Si * Si_energy_mlp - n_Ge * Ge_energy_mlp ) / len(atoms)
#     formation_energies_mlp_gpaw.append(formation_energy_mlp_gpaw)
#     energies_mlp_gpaw.append(atoms.get_potential_energy()/len(atoms))
#     print(n_Si, n_Ge, formation_energy_gpaw, formation_energy_mlp_gpaw, atoms.get_potential_energy()/len(atoms))


import matplotlib.pyplot as plt

## Write energies to a file
with open("Si-Ge_energies_comparison.txt", "w") as f:
    f.write("# DFT_Energy_per_atom MACE-MPA-0_Energy_per_atom CHGNet_Energy_per_atom\n")
    for dft_energy, mace_energy, chgnet_energy in zip(energies_dft, energies_mace, energies_chgnet):
        f.write(f"{dft_energy} {mace_energy} {chgnet_energy}\n")
    # for dft_energy, mace_energy in zip(energies_gpaw, energies_mlp_gpaw):
    #     f.write(f"{dft_energy} {mace_energy}\n")

with open("Si-Ge_formations_comparison.txt", "w") as f:
    f.write("# DFT_Formation_energy_per_atom MACE-MPA-0_Formation_energy_per_atom CHGNet_Formation_energy_per_atom\n")
    for dft_energy, mace_energy, chgnet_energy in zip(formation_energies_dft, formation_energies_mace, formation_energies_chgnet):
        f.write(f"{dft_energy} {mace_energy} {chgnet_energy}\n")
    # for dft_energy, mace_energy in zip(formation_energies_gpaw, formation_energies_mlp_gpaw):
    #     f.write(f"{dft_energy} {mace_energy}\n")

## Plotting the energies
plt.figure(figsize=(10, 6))
# plt.scatter(energies_dft, energies_mlp, c='blue', marker='o', label=label+' vs DFT')
plt.scatter(formation_energies_dft, formation_energies_mace, c='red', marker='o', label='MACE-MPA-0 vs DFT')
plt.scatter(formation_energies_dft, formation_energies_chgnet, c='blue', marker='o', label='CHGNet vs DFT')
plt.plot([min(formation_energies_dft), max(formation_energies_dft)], [min(formation_energies_dft), max(formation_energies_dft)], 'k--', label='x = y')
plt.xlabel('DFT Energy (eV/atom)')
plt.ylabel('MLIP Formation Energy (eV/atom)')
plt.legend()
plt.show()
